"""Airzone Cloud API Aidoo device."""

from __future__ import annotations

from typing import Any

from .common import SpeedType
from .const import (
    API_MODE,
    API_NAME,
    API_POWER,
    API_SETPOINT,
    API_SPEED_CONF,
    API_SPEED_TYPE,
    API_SPEED_VALUES,
    API_VALUE,
    AZD_SPEED,
    AZD_SPEED_TYPE,
    AZD_SPEEDS,
)
from .entity import EntityUpdate
from .hvac import HVAC


class Aidoo(HVAC):
    """Airzone Cloud Aidoo device."""

    def __init__(self, inst_id: str, ws_id: str, device_data: dict[str, Any]):
        """Airzone Cloud Aidoo device init."""
        super().__init__(inst_id, ws_id, device_data)

        self.speed: int | None = None
        self.speeds: dict[int, int] = {}
        self.speed_type: SpeedType | None = None

        if API_NAME in device_data:
            self.name = str(device_data[API_NAME])
        else:
            self.name = f"Aidoo {ws_id}"

    def data(self) -> dict[str, Any]:
        """Return Aidoo device data."""
        data = super().data()

        speed = self.get_speed()
        if speed is not None:
            data[AZD_SPEED] = speed

        speeds = self.get_speeds()
        if speeds is not None:
            data[AZD_SPEEDS] = speeds

        speed_type = self.get_speed_type()
        if speed_type is not None:
            data[AZD_SPEED_TYPE] = speed_type

        return data

    def get_speed(self) -> int | None:
        """Return Aidoo speed."""
        return self.speed

    def get_speeds(self) -> dict[int, int] | None:
        """Return Aidoo speeds."""
        if len(self.speeds) > 0:
            return self.speeds
        return None

    def get_speed_type(self) -> SpeedType | None:
        """Return Aidoo speed type."""
        return self.speed_type

    def set_speed(self, speed: int) -> None:
        """Set Aidoo speed."""
        if self.speed is not None:
            self.speed = speed

    def set_param(self, param: str, data: dict[str, Any]) -> None:
        """Update device parameter from API request."""
        if param == API_MODE:
            self.set_mode(data[API_VALUE])
        elif param == API_POWER:
            self.set_power(data[API_VALUE])
        elif param == API_SETPOINT:
            self.set_setpoint(data[API_VALUE])
        elif param == API_SPEED_CONF:
            self.set_speed(data[API_VALUE])

    def update_data(self, update: EntityUpdate) -> None:
        """Update Aidoo data."""
        super().update_data(update)

        data = update.get_data()

        speed = data.get(API_SPEED_CONF)
        if speed is not None:
            self.speed = int(speed)

        speed_type = data.get(API_SPEED_TYPE)
        if speed_type is not None:
            self.speed_type = SpeedType(speed_type)

        speeds_values: list[int] | None = data.get(API_SPEED_VALUES)
        if speeds_values is not None:
            speeds: dict[int, int] = {}

            if 0 in speeds_values:
                speeds[0] = 0

            speed_count = 1
            speeds_values.sort()
            for speed_value in speeds_values:
                if speed_value > 0:
                    speeds[speed_count] = int(speed_value)
                    speed_count += 1

            self.speeds = speeds



"""Airzone Cloud API HVAC device."""

from __future__ import annotations

from typing import Any

from .common import AirQualityMode, OperationAction, OperationMode
from .const import (
    API_ACTIVE,
    API_AQ_ACTIVE,
    API_AQ_MODE_CONF,
    API_AQ_MODE_VALUES,
    API_CELSIUS,
    API_DEFAULT_TEMP_STEP,
    API_HUMIDITY,
    API_LOCAL_TEMP,
    API_POWER,
    API_RANGE_MAX_AIR,
    API_RANGE_MIN_AIR,
    API_RANGE_SP_MAX_AUTO_AIR,
    API_RANGE_SP_MAX_COOL_AIR,
    API_RANGE_SP_MAX_DRY_AIR,
    API_RANGE_SP_MAX_EMERHEAT_AIR,
    API_RANGE_SP_MAX_HOT_AIR,
    API_RANGE_SP_MAX_STOP_AIR,
    API_RANGE_SP_MAX_VENT_AIR,
    API_RANGE_SP_MIN_AUTO_AIR,
    API_RANGE_SP_MIN_COOL_AIR,
    API_RANGE_SP_MIN_DRY_AIR,
    API_RANGE_SP_MIN_EMERHEAT_AIR,
    API_RANGE_SP_MIN_HOT_AIR,
    API_RANGE_SP_MIN_STOP_AIR,
    API_RANGE_SP_MIN_VENT_AIR,
    API_SP_AIR_COOL,
    API_SP_AIR_DRY,
    API_SP_AIR_HEAT,
    API_SP_AIR_STOP,
    API_SP_AIR_VENT,
    API_STEP,
    AZD_ACTION,
    AZD_ACTIVE,
    AZD_AQ_ACTIVE,
    AZD_AQ_MODE_CONF,
    AZD_AQ_MODE_VALUES,
    AZD_HUMIDITY,
    AZD_POWER,
    AZD_TEMP,
    AZD_TEMP_SET,
    AZD_TEMP_SET_AUTO_AIR,
    AZD_TEMP_SET_COOL_AIR,
    AZD_TEMP_SET_DRY_AIR,
    AZD_TEMP_SET_HOT_AIR,
    AZD_TEMP_SET_MAX,
    AZD_TEMP_SET_MAX_AUTO_AIR,
    AZD_TEMP_SET_MAX_COOL_AIR,
    AZD_TEMP_SET_MAX_DRY_AIR,
    AZD_TEMP_SET_MAX_EMERHEAT_AIR,
    AZD_TEMP_SET_MAX_HOT_AIR,
    AZD_TEMP_SET_MAX_STOP_AIR,
    AZD_TEMP_SET_MAX_VENT_AIR,
    AZD_TEMP_SET_MIN,
    AZD_TEMP_SET_MIN_AUTO_AIR,
    AZD_TEMP_SET_MIN_COOL_AIR,
    AZD_TEMP_SET_MIN_DRY_AIR,
    AZD_TEMP_SET_MIN_EMERHEAT_AIR,
    AZD_TEMP_SET_MIN_HOT_AIR,
    AZD_TEMP_SET_MIN_STOP_AIR,
    AZD_TEMP_SET_MIN_VENT_AIR,
    AZD_TEMP_SET_STOP_AIR,
    AZD_TEMP_SET_VENT_AIR,
    AZD_TEMP_STEP,
)
from .device import Device
from .entity import EntityUpdate, UpdateType


class HVAC(Device):
    """Airzone Cloud HVAC device."""

    def __init__(self, inst_id: str, ws_id: str, device_data: dict[str, Any]):
        """Airzone Cloud HVAC device init."""
        super().__init__(inst_id, ws_id, device_data)

        self.active: bool | None = None
        self.aq_active: bool | None = None
        self.aq_mode_conf: AirQualityMode | None = None
        self.aq_mode_values: list[AirQualityMode] | None = None
        self.humidity: int | None = None
        self.name: str = "HVAC"
        self.power: bool | None = None
        self.temp_set_max: float | None = None
        self.temp_set_max_auto_air: float | None = None
        self.temp_set_max_cool_air: float | None = None
        self.temp_set_max_dry_air: float | None = None
        self.temp_set_max_emerheat_air: float | None = None
        self.temp_set_max_hot_air: float | None = None
        self.temp_set_max_stop_air: float | None = None
        self.temp_set_max_vent_air: float | None = None
        self.temp_set_min: float | None = None
        self.temp_set_min_auto_air: float | None = None
        self.temp_set_min_cool_air: float | None = None
        self.temp_set_min_dry_air: float | None = None
        self.temp_set_min_emerheat_air: float | None = None
        self.temp_set_min_hot_air: float | None = None
        self.temp_set_min_stop_air: float | None = None
        self.temp_set_min_vent_air: float | None = None
        self.temp_set_auto_air: float | None = None
        self.temp_set_cool_air: float | None = None
        self.temp_set_dry_air: float | None = None
        self.temp_set_hot_air: float | None = None
        self.temp_set_stop_air: float | None = None
        self.temp_set_vent_air: float | None = None
        self.temp: float | None = None
        self.temp_step: float | None = None

    def data(self) -> dict[str, Any]:
        """Return HVAC device data."""
        data = super().data()

        data[AZD_ACTION] = self.get_action()
        data[AZD_ACTIVE] = self.get_active()
        data[AZD_POWER] = self.get_power()
        data[AZD_TEMP] = self.get_temperature()
        data[AZD_TEMP_STEP] = self.get_temp_step()

        aq_active = self.get_aq_active()
        if aq_active is not None:
            data[AZD_AQ_ACTIVE] = aq_active

        aq_mode_conf = self.get_aq_mode_conf()
        if aq_mode_conf is not None:
            data[AZD_AQ_MODE_CONF] = aq_mode_conf

        aq_mode_values = self.get_aq_mode_values()
        if aq_mode_values is not None:
            data[AZD_AQ_MODE_VALUES] = aq_mode_values

        humidity = self.get_humidity()
        if humidity is not None:
            data[AZD_HUMIDITY] = humidity

        temp_set_max = self.get_temp_set_max()
        if temp_set_max is not None:
            data[AZD_TEMP_SET_MAX] = temp_set_max
        temp_set_max_auto_air = self.get_temp_set_max_auto_air()
        if temp_set_max_auto_air is not None:
            data[AZD_TEMP_SET_MAX_AUTO_AIR] = temp_set_max_auto_air
        temp_set_max_cool_air = self.get_temp_set_max_cool_air()
        if temp_set_max_cool_air is not None:
            data[AZD_TEMP_SET_MAX_COOL_AIR] = temp_set_max_cool_air
        temp_set_max_dry_air = self.get_temp_set_max_dry_air()
        if temp_set_max_dry_air is not None:
            data[AZD_TEMP_SET_MAX_DRY_AIR] = temp_set_max_dry_air
        temp_set_max_emerheat_air = self.get_temp_set_max_emerheat_air()
        if temp_set_max_emerheat_air is not None:
            data[AZD_TEMP_SET_MAX_EMERHEAT_AIR] = temp_set_max_emerheat_air
        temp_set_max_hot_air = self.get_temp_set_max_hot_air()
        if temp_set_max_hot_air is not None:
            data[AZD_TEMP_SET_MAX_HOT_AIR] = temp_set_max_hot_air
        temp_set_max_stop_air = self.get_temp_set_max_stop_air()
        if temp_set_max_stop_air is not None:
            data[AZD_TEMP_SET_MAX_STOP_AIR] = temp_set_max_stop_air
        temp_set_max_vent_air = self.get_temp_set_max_vent_air()
        if temp_set_max_vent_air is not None:
            data[AZD_TEMP_SET_MAX_VENT_AIR] = temp_set_max_vent_air

        temp_set_min = self.get_temp_set_min()
        if temp_set_min is not None:
            data[AZD_TEMP_SET_MIN] = temp_set_min
        temp_set_min_auto_air = self.get_temp_set_min_auto_air()
        if temp_set_min_auto_air is not None:
            data[AZD_TEMP_SET_MIN_AUTO_AIR] = temp_set_min_auto_air
        temp_set_min_cool_air = self.get_temp_set_min_cool_air()
        if temp_set_min_cool_air is not None:
            data[AZD_TEMP_SET_MIN_COOL_AIR] = temp_set_min_cool_air
        temp_set_min_dry_air = self.get_temp_set_min_dry_air()
        if temp_set_min_dry_air is not None:
            data[AZD_TEMP_SET_MIN_DRY_AIR] = temp_set_min_dry_air
        temp_set_min_emerheat_air = self.get_temp_set_min_emerheat_air()
        if temp_set_min_emerheat_air is not None:
            data[AZD_TEMP_SET_MIN_EMERHEAT_AIR] = temp_set_min_emerheat_air
        temp_set_min_hot_air = self.get_temp_set_min_hot_air()
        if temp_set_min_hot_air is not None:
            data[AZD_TEMP_SET_MIN_HOT_AIR] = temp_set_min_hot_air
        temp_set_min_stop_air = self.get_temp_set_min_stop_air()
        if temp_set_min_stop_air is not None:
            data[AZD_TEMP_SET_MIN_STOP_AIR] = temp_set_min_stop_air
        temp_set_min_vent_air = self.get_temp_set_min_vent_air()
        if temp_set_min_vent_air is not None:
            data[AZD_TEMP_SET_MIN_VENT_AIR] = temp_set_min_vent_air

        temp_set = self.get_temp_set()
        if temp_set is not None:
            data[AZD_TEMP_SET] = temp_set
        temp_set_auto_air = self.get_temp_set_auto_air()
        if temp_set_auto_air is not None:
            data[AZD_TEMP_SET_AUTO_AIR] = temp_set_auto_air
        temp_set_cool_air = self.get_temp_set_cool_air()
        if temp_set_cool_air is not None:
            data[AZD_TEMP_SET_COOL_AIR] = temp_set_cool_air
        temp_set_dry_air = self.get_temp_set_dry_air()
        if temp_set_dry_air is not None:
            data[AZD_TEMP_SET_DRY_AIR] = temp_set_dry_air
        temp_set_hot_air = self.get_temp_set_hot_air()
        if temp_set_hot_air is not None:
            data[AZD_TEMP_SET_HOT_AIR] = temp_set_hot_air
        temp_set_stop_air = self.get_temp_set_stop_air()
        if temp_set_stop_air is not None:
            data[AZD_TEMP_SET_STOP_AIR] = temp_set_stop_air
        temp_set_vent_air = self.get_temp_set_vent_air()
        if temp_set_vent_air is not None:
            data[AZD_TEMP_SET_VENT_AIR] = temp_set_vent_air

        return data

    def get_action(self) -> OperationAction:
        """Return HVAC action."""
        if self.get_power():
            if self.get_active():
                mode = self.get_mode() or OperationMode.STOP
                if mode.is_cool():
                    action = OperationAction.COOLING
                elif mode.is_heat():
                    action = OperationAction.HEATING
                elif mode.is_vent():
                    action = OperationAction.FAN
                elif mode.is_dry():
                    action = OperationAction.DRYING
                elif mode.is_auto():
                    action = self.get_action_auto()
                else:
                    action = OperationAction.OFF
            else:
                action = OperationAction.IDLE
        else:
            action = OperationAction.OFF
        return action

    def get_action_auto(self) -> OperationAction:
        """Return current Auto action."""
        mode = self.get_mode_auto()
        if mode is not None:
            if mode.is_cool():
                action = OperationAction.COOLING
            elif mode.is_heat():
                action = OperationAction.HEATING
            elif mode.is_vent():
                action = OperationAction.FAN
            elif mode.is_dry():
                action = OperationAction.DRYING
            else:
                action = OperationAction.IDLE
        else:
            action = OperationAction.IDLE
        return action

    def get_active(self) -> bool | None:
        """Return HVAC device active status."""
        return self.active

    def get_aq_active(self) -> bool | None:
        """Return HVAC device Air Quality active status."""
        return self.aq_active

    def get_aq_mode_conf(self) -> AirQualityMode | None:
        """Return HVAC device Air Quality mode conf."""
        return self.aq_mode_conf

    def get_aq_mode_values(self) -> list[AirQualityMode] | None:
        """Return HVAC device Air Quality mode values."""
        if self.aq_mode_values is not None and len(self.aq_mode_values) > 0:
            return self.aq_mode_values
        return None

    def get_humidity(self) -> int | None:
        """Return HVAC device humidity."""
        return self.humidity

    def get_power(self) -> bool | None:
        """Return HVAC device power."""
        return self.power

    def get_temperature(self) -> float | None:
        """Return HVAC device temperature."""
        if self.temp is not None:
            return round(self.temp, 1)
        return None

    def get_temp_set(self) -> float | None:
        """Return HVAC device temperature setpoint."""
        temp_set: float | None = None
        mode = self.get_mode()
        if mode is not None:
            if mode.is_auto():
                temp_set = self.get_temp_set_auto_air()
            elif mode.is_cool():
                temp_set = self.get_temp_set_cool_air()
            elif mode.is_dry():
                temp_set = self.get_temp_set_dry_air()
            elif mode.is_heat():
                temp_set = self.get_temp_set_hot_air()
            elif mode.is_stop():
                temp_set = self.get_temp_set_stop_air()
            elif mode.is_vent():
                temp_set = self.get_temp_set_vent_air()
        return temp_set

    def get_temp_set_auto_air(self) -> float | None:
        """Return HVAC device setpoint for Auto Air."""
        if self.temp_set_auto_air is not None:
            return round(self.temp_set_auto_air, 1)
        return None

    def get_temp_set_cool_air(self) -> float | None:
        """Return HVAC device setpoint for Cool Air."""
        if self.temp_set_cool_air is not None:
            return round(self.temp_set_cool_air, 1)
        return None

    def get_temp_set_dry_air(self) -> float | None:
        """Return HVAC device setpoint for Dry Air."""
        if self.temp_set_dry_air is not None:
            return round(self.temp_set_dry_air, 1)
        return None

    def get_temp_set_hot_air(self) -> float | None:
        """Return HVAC device setpoint for Heat Air."""
        if self.temp_set_hot_air is not None:
            return round(self.temp_set_hot_air, 1)
        return None

    def get_temp_set_stop_air(self) -> float | None:
        """Return HVAC device setpoint for Stop Air."""
        if self.temp_set_stop_air is not None:
            return round(self.temp_set_stop_air, 1)
        return None

    def get_temp_set_vent_air(self) -> float | None:
        """Return HVAC device setpoint for Ventilation Air."""
        if self.temp_set_vent_air is not None:
            return round(self.temp_set_vent_air, 1)
        return None

    def get_temp_set_max(self) -> float | None:
        """Return HVAC device max setpoint."""
        max_temp: float | None = None
        temps = [
            self.temp_set_max,
            self.get_temp_set_max_auto_air(),
            self.get_temp_set_max_cool_air(),
            self.get_temp_set_max_dry_air(),
            self.get_temp_set_max_emerheat_air(),
            self.get_temp_set_max_hot_air(),
            self.get_temp_set_max_stop_air(),
            self.get_temp_set_max_vent_air(),
        ]
        for temp in temps:
            if temp is not None:
                if max_temp is None or temp > max_temp:
                    max_temp = temp
        return max_temp

    def get_temp_set_max_auto_air(self) -> float | None:
        """Return HVAC device max setpoint for Auto Air."""
        if self.temp_set_max_auto_air is not None:
            return round(self.temp_set_max_auto_air, 1)
        return None

    def get_temp_set_max_cool_air(self) -> float | None:
        """Return HVAC device max setpoint for Cool Air."""
        if self.temp_set_max_cool_air is not None:
            return round(self.temp_set_max_cool_air, 1)
        return None

    def get_temp_set_max_dry_air(self) -> float | None:
        """Return HVAC device max setpoint for Dry Air."""
        if self.temp_set_max_dry_air is not None:
            return round(self.temp_set_max_dry_air, 1)
        return None

    def get_temp_set_max_emerheat_air(self) -> float | None:
        """Return HVAC device max setpoint for Emergency Heat Air."""
        if self.temp_set_max_emerheat_air is not None:
            return round(self.temp_set_max_emerheat_air, 1)
        return None

    def get_temp_set_max_hot_air(self) -> float | None:
        """Return HVAC device max setpoint for Heat Air."""
        if self.temp_set_max_hot_air is not None:
            return round(self.temp_set_max_hot_air, 1)
        return None

    def get_temp_set_max_stop_air(self) -> float | None:
        """Return HVAC device max setpoint for Stop Air."""
        if self.temp_set_max_stop_air is not None:
            return round(self.temp_set_max_stop_air, 1)
        return None

    def get_temp_set_max_vent_air(self) -> float | None:
        """Return HVAC device max setpoint for Ventilation Air."""
        if self.temp_set_max_vent_air is not None:
            return round(self.temp_set_max_vent_air, 1)
        return None

    def get_temp_set_min(self) -> float | None:
        """Return HVAC device min setpoint."""
        min_temp: float | None = None
        temps = [
            self.temp_set_min,
            self.get_temp_set_min_auto_air(),
            self.get_temp_set_min_cool_air(),
            self.get_temp_set_min_dry_air(),
            self.get_temp_set_min_emerheat_air(),
            self.get_temp_set_min_hot_air(),
            self.get_temp_set_min_stop_air(),
            self.get_temp_set_min_vent_air(),
        ]
        for temp in temps:
            if temp is not None:
                if min_temp is None or temp < min_temp:
                    min_temp = temp
        return min_temp

    def get_temp_set_min_auto_air(self) -> float | None:
        """Return HVAC device min setpoint for Auto Air."""
        if self.temp_set_min_auto_air:
            return round(self.temp_set_min_auto_air, 1)
        return None

    def get_temp_set_min_cool_air(self) -> float | None:
        """Return HVAC device min setpoint for Cool Air."""
        if self.temp_set_min_cool_air:
            return round(self.temp_set_min_cool_air, 1)
        return None

    def get_temp_set_min_dry_air(self) -> float | None:
        """Return HVAC device min setpoint for Dry Air."""
        if self.temp_set_min_dry_air is not None:
            return round(self.temp_set_min_dry_air, 1)
        return None

    def get_temp_set_min_emerheat_air(self) -> float | None:
        """Return HVAC device min setpoint for Emergency Heat Air."""
        if self.temp_set_min_emerheat_air is not None:
            return round(self.temp_set_min_emerheat_air, 1)
        return None

    def get_temp_set_min_hot_air(self) -> float | None:
        """Return HVAC device min setpoint for Hot Air."""
        if self.temp_set_min_hot_air is not None:
            return round(self.temp_set_min_hot_air, 1)
        return None

    def get_temp_set_min_stop_air(self) -> float | None:
        """Return HVAC device min setpoint for Stop Air."""
        if self.temp_set_min_stop_air is not None:
            return round(self.temp_set_min_stop_air, 1)
        return None

    def get_temp_set_min_vent_air(self) -> float | None:
        """Return HVAC device min setpoint for Ventilation Air."""
        if self.temp_set_min_vent_air is not None:
            return round(self.temp_set_min_vent_air, 1)
        return None

    def get_temp_step(self) -> float | None:
        """Return HVAC device temperature step."""
        if self.temp_step is not None:
            return round(self.temp_step, 1)
        return API_DEFAULT_TEMP_STEP

    def set_aq_mode(self, aq_mode: AirQualityMode) -> None:
        """Set HVAC Air Quality mode."""
        self.aq_mode_conf = aq_mode

    def set_power(self, power: bool) -> None:
        """Set HVAC power."""
        self.power = power

    def set_setpoint(self, setpoint: float) -> None:
        """Set HVAC setpoint."""
        if self.temp_set_auto_air is not None:
            self.temp_set_auto_air = setpoint
        if self.temp_set_cool_air is not None:
            self.temp_set_cool_air = setpoint
        if self.temp_set_dry_air is not None:
            self.temp_set_dry_air = setpoint
        if self.temp_set_hot_air is not None:
            self.temp_set_hot_air = setpoint
        if self.temp_set_stop_air is not None:
            self.temp_set_stop_air = setpoint
        if self.temp_set_vent_air is not None:
            self.temp_set_vent_air = setpoint

    def update_data(self, update: EntityUpdate) -> None:
        """Update HVAC device data."""
        super().update_data(update)

        data = update.get_data()

        if API_ACTIVE in data:
            active = data.get(API_ACTIVE)
            if active is not None:
                self.active = bool(active)
            else:
                # API sends active as null instead of False
                self.active = False
        else:
            if update.get_type() != UpdateType.WS_PARTIAL:
                self.active = None

        aq_active = data.get(API_AQ_ACTIVE)
        if aq_active is not None:
            self.aq_active = bool(aq_active)

        aq_mode_conf = data.get(API_AQ_MODE_CONF)
        if aq_mode_conf is not None:
            self.aq_mode_conf = AirQualityMode(aq_mode_conf)

        aq_mode_values = data.get(API_AQ_MODE_VALUES)
        if aq_mode_values is not None:
            self.aq_mode_values = []
            for aq_mode_value in aq_mode_values:
                self.aq_mode_values += [AirQualityMode(aq_mode_value)]

        humidity = data.get(API_HUMIDITY)
        if humidity is not None:
            self.humidity = int(humidity)

        local_temp = data.get(API_LOCAL_TEMP)
        if local_temp is not None:
            if API_CELSIUS in local_temp:
                self.temp = float(local_temp[API_CELSIUS])

        power = data.get(API_POWER)
        if power is not None:
            self.power = bool(power)

        range_max_air = data.get(API_RANGE_MAX_AIR)
        if range_max_air is not None:
            if API_CELSIUS in range_max_air:
                self.temp_set_max = float(range_max_air[API_CELSIUS])
        range_sp_max_auto_air = data.get(API_RANGE_SP_MAX_AUTO_AIR)
        if range_sp_max_auto_air is not None:
            if API_CELSIUS in range_sp_max_auto_air:
                self.temp_set_max_auto_air = float(range_sp_max_auto_air[API_CELSIUS])
        range_sp_max_cool_air = data.get(API_RANGE_SP_MAX_COOL_AIR)
        if range_sp_max_cool_air is not None:
            if API_CELSIUS in range_sp_max_cool_air:
                self.temp_set_max_cool_air = float(range_sp_max_cool_air[API_CELSIUS])
        range_sp_max_dry_air = data.get(API_RANGE_SP_MAX_DRY_AIR)
        if range_sp_max_dry_air is not None:
            if API_CELSIUS in range_sp_max_dry_air:
                self.temp_set_max_dry_air = float(range_sp_max_dry_air[API_CELSIUS])
        range_sp_max_emerheat_air = data.get(API_RANGE_SP_MAX_EMERHEAT_AIR)
        if range_sp_max_emerheat_air is not None:
            if API_CELSIUS in range_sp_max_emerheat_air:
                self.temp_set_max_emerheat_air = float(
                    range_sp_max_emerheat_air[API_CELSIUS]
                )
        range_sp_max_hot_air = data.get(API_RANGE_SP_MAX_HOT_AIR)
        if range_sp_max_hot_air is not None:
            if API_CELSIUS in range_sp_max_hot_air:
                self.temp_set_max_hot_air = float(range_sp_max_hot_air[API_CELSIUS])
        range_sp_max_stop_air = data.get(API_RANGE_SP_MAX_STOP_AIR)
        if range_sp_max_stop_air is not None:
            if API_CELSIUS in range_sp_max_stop_air:
                self.temp_set_max_stop_air = float(range_sp_max_stop_air[API_CELSIUS])
        range_sp_max_vent_air = data.get(API_RANGE_SP_MAX_VENT_AIR)
        if range_sp_max_vent_air is not None:
            if API_CELSIUS in range_sp_max_vent_air:
                self.temp_set_max_vent_air = float(range_sp_max_vent_air[API_CELSIUS])

        range_min_air = data.get(API_RANGE_MIN_AIR)
        if range_min_air is not None:
            if API_CELSIUS in range_min_air:
                self.temp_set_min = float(range_min_air[API_CELSIUS])
        range_sp_min_auto_air = data.get(API_RANGE_SP_MIN_AUTO_AIR)
        if range_sp_min_auto_air is not None:
            if API_CELSIUS in range_sp_min_auto_air:
                self.temp_set_min_auto_air = float(range_sp_min_auto_air[API_CELSIUS])
        range_sp_min_cool_air = data.get(API_RANGE_SP_MIN_COOL_AIR)
        if range_sp_min_cool_air is not None:
            if API_CELSIUS in range_sp_min_cool_air:
                self.temp_set_min_cool_air = float(range_sp_min_cool_air[API_CELSIUS])
        range_sp_min_dry_air = data.get(API_RANGE_SP_MIN_DRY_AIR)
        if range_sp_min_dry_air is not None:
            if API_CELSIUS in range_sp_min_dry_air:
                self.temp_set_min_dry_air = float(range_sp_min_dry_air[API_CELSIUS])
        range_sp_min_emerheat_air = data.get(API_RANGE_SP_MIN_EMERHEAT_AIR)
        if range_sp_min_emerheat_air is not None:
            if API_CELSIUS in range_sp_min_emerheat_air:
                self.temp_set_min_emerheat_air = float(
                    range_sp_min_emerheat_air[API_CELSIUS]
                )
        range_sp_min_hot_air = data.get(API_RANGE_SP_MIN_HOT_AIR)
        if range_sp_min_hot_air is not None:
            if API_CELSIUS in range_sp_min_hot_air:
                self.temp_set_min_hot_air = float(range_sp_min_hot_air[API_CELSIUS])
        range_sp_min_stop_air = data.get(API_RANGE_SP_MIN_STOP_AIR)
        if range_sp_min_stop_air is not None:
            if API_CELSIUS in range_sp_min_stop_air:
                self.temp_set_min_stop_air = float(range_sp_min_stop_air[API_CELSIUS])
        range_sp_min_vent_air = data.get(API_RANGE_SP_MIN_VENT_AIR)
        if range_sp_min_vent_air is not None:
            if API_CELSIUS in range_sp_min_vent_air:
                self.temp_set_min_vent_air = float(range_sp_min_vent_air[API_CELSIUS])

        sp_air_cool = data.get(API_SP_AIR_COOL)
        if sp_air_cool is not None:
            if API_CELSIUS in sp_air_cool:
                self.temp_set_cool_air = float(sp_air_cool[API_CELSIUS])
        sp_air_dry = data.get(API_SP_AIR_DRY)
        if sp_air_dry is not None:
            if API_CELSIUS in sp_air_dry:
                self.temp_set_dry_air = float(data[API_SP_AIR_DRY][API_CELSIUS])
        sp_air_heat = data.get(API_SP_AIR_HEAT)
        if sp_air_heat is not None:
            if API_CELSIUS in sp_air_heat:
                self.temp_set_hot_air = float(sp_air_heat[API_CELSIUS])
        sp_air_stop = data.get(API_SP_AIR_STOP)
        if sp_air_stop is not None:
            if API_CELSIUS in sp_air_stop:
                self.temp_set_stop_air = float(sp_air_stop[API_CELSIUS])
        sp_air_vent = data.get(API_SP_AIR_VENT)
        if sp_air_vent is not None:
            if API_CELSIUS in sp_air_vent:
                self.temp_set_vent_air = float(sp_air_vent[API_CELSIUS])

        step = data.get(API_STEP)
        if step is not None:
            if API_CELSIUS in step:
                self.temp_step = float(step[API_CELSIUS])



"""Airzone Cloud API System."""

from __future__ import annotations

from typing import TYPE_CHECKING, Any

from .const import API_MODE, API_SYSTEM_NUMBER, API_VALUE, AZD_SYSTEM
from .device import Device

if TYPE_CHECKING:
    from .zone import Zone


class System(Device):
    """Airzone Cloud System device."""

    def __init__(self, inst_id: str, ws_id: str, device_data: dict[str, Any]):
        """Airzone Cloud System device init."""
        super().__init__(inst_id, ws_id, device_data)

        self.zones: dict[str, Zone] = {}

        sub_data = self.sub_data(device_data)
        self.system_number = int(sub_data[API_SYSTEM_NUMBER])

        self.name = f"System {self.system_number}"

    def data(self) -> dict[str, Any]:
        """Return System device data."""
        data = super().data()

        data[AZD_SYSTEM] = self.get_system_num()

        return data

    def add_zone(self, zone: Zone) -> None:
        """Add System zone."""
        zone_id = zone.get_id()
        if zone_id not in self.zones:
            self.zones[zone_id] = zone

    def get_system_num(self) -> int:
        """Return System number."""
        return self.system_number

    def set_param(self, param: str, data: dict[str, Any]) -> None:
        """Update System parameter from API request."""

        if param == API_MODE:
            self.set_mode(data[API_VALUE])
            for zone in self.zones.values():
                zone.set_mode(data[API_VALUE])



"""Airzone Cloud API Zone device."""

from __future__ import annotations

from typing import TYPE_CHECKING, Any

from aioairzone_cloud.common import OperationMode

from .const import (
    API_AQ_MODE_CONF,
    API_MODE,
    API_MODE_AVAIL,
    API_NAME,
    API_POWER,
    API_SETPOINT,
    API_SYSTEM_NUMBER,
    API_VALUE,
    API_ZONE_NUMBER,
    AZD_MASTER,
    AZD_SYSTEM,
    AZD_SYSTEM_ID,
    AZD_ZONE,
)
from .entity import EntityUpdate, UpdateType
from .hvac import HVAC

if TYPE_CHECKING:
    from .system import System


class Zone(HVAC):
    """Airzone Cloud Zone device."""

    def __init__(self, inst_id: str, ws_id: str, device_data: dict[str, Any]):
        """Airzone Cloud Zone device init."""
        super().__init__(inst_id, ws_id, device_data)

        self.master: bool | None = None
        self.system: System | None = None

        sub_data = self.sub_data(device_data)
        self.system_number = int(sub_data[API_SYSTEM_NUMBER])
        self.zone_number = int(sub_data[API_ZONE_NUMBER])

        if API_NAME in device_data:
            self.name = str(device_data[API_NAME])
        else:
            self.name = f"Zone {self.system_number}:{self.zone_number}"

    def data(self) -> dict[str, Any]:
        """Return Zone data."""
        data = super().data()

        data[AZD_MASTER] = self.get_master()
        data[AZD_SYSTEM] = self.get_system_num()
        data[AZD_ZONE] = self.get_zone()

        system_id = self.get_system_id()
        if system_id is not None:
            data[AZD_SYSTEM_ID] = system_id

        return data

    def get_master(self) -> bool:
        """Return Zone device master/slave."""
        return self.master or False

    def get_system(self) -> System | None:
        """Return Zone device System."""
        if self.system is not None:
            return self.system
        return None

    def get_system_id(self) -> str | None:
        """Return Zone device System ID."""
        if self.system is not None:
            return self.system.get_id()
        return None

    def get_system_num(self) -> int:
        """Return System number."""
        return self.system_number

    def get_zone(self) -> int:
        """Return Zone number."""
        return self.zone_number

    def set_modes(self, modes: list[OperationMode]) -> None:
        """Set slave zone modes."""
        self.modes = modes

    def set_param(self, param: str, data: dict[str, Any]) -> None:
        """Update device parameter from API request."""

        if param == API_AQ_MODE_CONF:
            self.set_aq_mode(data[API_VALUE])
        elif param == API_MODE:
            if self.system is not None:
                self.system.set_param(param, data)
            else:
                self.set_mode(data[API_VALUE])
        elif param == API_POWER:
            self.set_power(data[API_VALUE])
        elif param == API_SETPOINT:
            self.set_setpoint(data[API_VALUE])

    def set_system(self, system: System) -> None:
        """Set System."""
        self.system = system

    def update_data(self, update: EntityUpdate) -> None:
        """Update Zone data."""
        super().update_data(update)

        if update.get_type() == UpdateType.API_FULL:
            data = update.get_data()

            if API_MODE_AVAIL in data:
                self.master = len(data[API_MODE_AVAIL]) > 0
            else:
                self.master = None



"""Airzone Cloud API Device."""

from __future__ import annotations

from abc import ABC, abstractmethod
from asyncio import Lock
from datetime import datetime
from enum import IntEnum
import logging
from typing import Any

from .const import WS_CHANGE, WS_STATUS

_LOGGER = logging.getLogger(__name__)


class UpdateType(IntEnum):
    """Airzone Cloud Update type."""

    API_FULL = 1
    WS_FULL = 2
    WS_PARTIAL = 3


class EntityUpdate:
    """Airzone Cloud Entity Update."""

    data: dict[str, Any]

    def __init__(self, _type: UpdateType, data: dict[str, Any]):
        """Airzone Cloud Update init."""
        self.datetime: datetime = datetime.now()
        self.data: dict[str, Any] = data
        self.type: UpdateType = _type

    def check_dt(self, dt: datetime) -> bool:
        """Check if Update data is newer than provided datetime."""
        return self.datetime >= dt

    def get_data(self) -> dict[str, Any]:
        """Get Entity Update data."""
        if self.type == UpdateType.WS_PARTIAL:
            change: dict[str, Any] = self.data.get(WS_CHANGE, {})
            status: dict[str, Any] = change.get(WS_STATUS, {})
            return status

        return self.data

    def get_datetime(self) -> datetime:
        """Get Entity Update datetime."""
        return self.datetime

    def get_type(self) -> UpdateType:
        """Get Entity Update type."""
        return self.type

    def set_data(self, data: dict[str, Any]) -> None:
        """Set Entity Update data."""
        self.data = data

    def __str__(self) -> str:
        """Return Entity Update string representation."""
        _data = self.get_data()
        _type = self.get_type()
        return f"{_type} with data={_data}"


class Entity(ABC):
    """Airzone Cloud Entity."""

    datetime: datetime
    id: str
    lock: Lock
    name: str

    def __init__(self) -> None:
        """Airzone Cloud Device init."""
        self.datetime: datetime = datetime.now()
        self.init: bool = False
        self.lock: Lock = Lock()

    @abstractmethod
    def data(self) -> dict[str, Any]:
        """Return Entity data."""

    def get_id(self) -> str:
        """Return Entity ID."""
        return self.id

    def get_init(self) -> bool:
        """Return Entity Init."""
        return self.init

    def get_name(self) -> str:
        """Return Entity name."""
        return self.name

    @abstractmethod
    def update_data(self, update: EntityUpdate) -> None:
        """Update Entity data."""

    async def update(self, update: EntityUpdate) -> None:
        """Update Entity."""
        newer: bool = update.check_dt(self.datetime)

        _LOGGER.debug(
            "%s[%s] update (newer=%s) update=%s",
            type(self).__name__,
            self.get_id(),
            newer,
            update,
        )

        if newer:
            async with self.lock:
                self.update_data(update)
                self.datetime = update.get_datetime()
                if update.get_type() == UpdateType.API_FULL:
                    self.init = True



"""Airzone Cloud Local API based device."""

from __future__ import annotations

import logging
from typing import Any

from .const import (
    API_CONFIG,
    API_CONNECTION_DATE,
    API_DISCONNECTION_DATE,
    API_IS_CONNECTED,
    API_STAT_AP_MAC,
    API_STAT_CHANNEL,
    API_STAT_QUALITY,
    API_STAT_RSSI,
    API_STAT_SSID,
    API_STATUS,
    API_WS_FW,
    API_WS_TYPE,
    AZD_AVAILABLE,
    AZD_CONNECTION_DATE,
    AZD_DISCONNECTION_DATE,
    AZD_FIRMWARE,
    AZD_ID,
    AZD_INSTALLATION,
    AZD_NAME,
    AZD_TYPE,
    AZD_WIFI_CHANNEL,
    AZD_WIFI_MAC,
    AZD_WIFI_QUALITY,
    AZD_WIFI_RSSI,
    AZD_WIFI_SSID,
)
from .entity import Entity, EntityUpdate

_LOGGER = logging.getLogger(__name__)


class WebServer(Entity):
    """Airzone Cloud WebServer."""

    def __init__(self, inst_id: str, ws_id: str):
        """Airzone Cloud WebServer init."""
        super().__init__()

        self.connection_date: str | None = None
        self.disconnection_date: str | None = None
        self.firmware: str | None = None
        self.id = ws_id
        self.installation_id = inst_id
        self.is_connected: bool = False
        self.name: str = f"WebServer {ws_id}"
        self.stat_quality: int | None = None
        self.type: str | None = None
        self.wifi_channel: int | None = None
        self.wifi_mac: str | None = None
        self.wifi_quality: int | None = None
        self.wifi_rssi: int | None = None
        self.wifi_ssid: str | None = None

    def update_data(self, update: EntityUpdate) -> None:
        """Update WebServer data."""
        data = update.get_data()

        ws_type = data.get(API_WS_TYPE)
        if ws_type is not None:
            self.type = str(ws_type)

        ws_config = data.get(API_CONFIG)
        if ws_config is not None:
            stat_ap_mac = ws_config.get(API_STAT_AP_MAC)
            if stat_ap_mac is not None:
                self.wifi_mac = str(stat_ap_mac)
            stat_channel = ws_config.get(API_STAT_CHANNEL)
            if stat_channel is not None:
                self.wifi_channel = int(stat_channel)
            stat_ssid = ws_config.get(API_STAT_SSID)
            if stat_ssid is not None:
                self.wifi_ssid = str(stat_ssid)
            ws_fw = ws_config.get(API_WS_FW)
            if ws_fw is not None:
                self.firmware = str(ws_fw)

        ws_status = data.get(API_STATUS)
        if ws_status is not None:
            connection_date = ws_status.get(API_CONNECTION_DATE)
            if connection_date is not None:
                self.connection_date = str(connection_date)
            disconnection_date = ws_status.get(API_DISCONNECTION_DATE)
            if disconnection_date is not None:
                self.disconnection_date = str(disconnection_date)
            is_connected = ws_status.get(API_IS_CONNECTED)
            if is_connected is not None:
                self.is_connected = bool(is_connected)
            stat_quality = ws_status.get(API_STAT_QUALITY)
            if stat_quality is not None:
                self.wifi_quality = int(stat_quality)
            stat_rssi = ws_status.get(API_STAT_RSSI)
            if stat_rssi is not None:
                self.wifi_rssi = int(stat_rssi)

    def data(self) -> dict[str, Any]:
        """Return WebServer data."""
        data: dict[str, Any] = {
            AZD_AVAILABLE: self.get_available(),
            AZD_CONNECTION_DATE: self.get_connection_date(),
            AZD_DISCONNECTION_DATE: self.get_disconnection_date(),
            AZD_FIRMWARE: self.get_firmware(),
            AZD_ID: self.get_id(),
            AZD_INSTALLATION: self.get_installation(),
            AZD_NAME: self.get_name(),
            AZD_TYPE: self.get_type(),
        }

        wifi_channel = self.get_wifi_channel()
        if wifi_channel is not None:
            data[AZD_WIFI_CHANNEL] = wifi_channel

        wifi_mac = self.get_wifi_mac()
        if wifi_mac is not None:
            data[AZD_WIFI_MAC] = wifi_mac

        wifi_quality = self.get_wifi_quality()
        if wifi_quality is not None:
            data[AZD_WIFI_QUALITY] = wifi_quality

        wifi_rssi = self.get_wifi_rssi()
        if wifi_rssi is not None:
            data[AZD_WIFI_RSSI] = wifi_rssi

        wifi_ssid = self.get_wifi_ssid()
        if wifi_ssid is not None:
            data[AZD_WIFI_SSID] = wifi_ssid

        return data

    def get_available(self) -> bool:
        """Return availability status."""
        return self.is_connected

    def get_connection_date(self) -> str | None:
        """Return connection date."""
        return self.connection_date

    def get_disconnection_date(self) -> str | None:
        """Return disconnection date."""
        return self.disconnection_date

    def get_firmware(self) -> str | None:
        """Return firmware version."""
        return self.firmware

    def get_id(self) -> str:
        """Return WebServer ID."""
        return self.id

    def get_installation(self) -> str:
        """Return installation ID."""
        return self.installation_id

    def get_name(self) -> str:
        """Return WebServer name."""
        return self.name

    def get_type(self) -> str | None:
        """Return WebServer type."""
        return self.type

    def get_wifi_channel(self) -> int | None:
        """Return WiFi channel."""
        return self.wifi_channel

    def get_wifi_mac(self) -> str | None:
        """Return WiFi Mac address."""
        return self.wifi_mac

    def get_wifi_quality(self) -> int | None:
        """Return WiFi signal quality."""
        return self.wifi_quality

    def get_wifi_rssi(self) -> int | None:
        """Return WiFi RSSI."""
        return self.wifi_rssi

    def get_wifi_ssid(self) -> str | None:
        """Return WiFi SSID."""
        return self.wifi_ssid



"""Airzone Cloud API Device."""

from __future__ import annotations

from abc import ABC
from collections import Counter
from typing import Any

from .aidoo import Aidoo
from .common import OperationAction, OperationMode
from .const import (
    API_DEFAULT_TEMP_STEP,
    API_PARAMS,
    API_VALUE,
    AZD_ACTION,
    AZD_ACTIVE,
    AZD_AIDOOS,
    AZD_AVAILABLE,
    AZD_HUMIDITY,
    AZD_ID,
    AZD_MODE,
    AZD_MODES,
    AZD_NAME,
    AZD_NUM_DEVICES,
    AZD_POWER,
    AZD_SYSTEMS,
    AZD_TEMP,
    AZD_TEMP_SET,
    AZD_TEMP_SET_MAX,
    AZD_TEMP_SET_MIN,
    AZD_TEMP_STEP,
    AZD_ZONES,
)
from .system import System
from .zone import Zone


class DeviceGroup(ABC):
    """Airzone Cloud DeviceGroup."""

    id: str
    name: str

    def __init__(self) -> None:
        """Airzone Cloud DeviceGroup init."""

        self.aidoos: dict[str, Aidoo] = {}
        self.systems: dict[str, System] = {}
        self.zones: dict[str, Zone] = {}

    def data(self) -> dict[str, Any]:
        """Return DeviceGroup data."""
        data: dict[str, Any] = {}

        if len(self.aidoos) > 0:
            data[AZD_AIDOOS] = list(self.aidoos)
        if len(self.systems) > 0:
            data[AZD_SYSTEMS] = list(self.systems)
        if len(self.zones) > 0:
            data[AZD_ZONES] = list(self.zones)

        data[AZD_ACTION] = self.get_action()
        data[AZD_ACTIVE] = self.get_active()
        data[AZD_AVAILABLE] = self.get_available()
        data[AZD_ID] = self.get_id()
        data[AZD_MODE] = self.get_mode()
        data[AZD_NAME] = self.get_name()
        data[AZD_NUM_DEVICES] = self.get_devices_num()
        data[AZD_POWER] = self.get_power()
        data[AZD_TEMP] = self.get_temperature()
        data[AZD_TEMP_STEP] = self.get_temp_step()

        humidity = self.get_humidity()
        if humidity is not None:
            data[AZD_HUMIDITY] = humidity

        modes = self.get_modes()
        if modes is not None:
            data[AZD_MODES] = modes

        temp_set_max = self.get_temp_set_max()
        if temp_set_max is not None:
            data[AZD_TEMP_SET_MAX] = temp_set_max

        temp_set_min = self.get_temp_set_min()
        if temp_set_min is not None:
            data[AZD_TEMP_SET_MIN] = temp_set_min

        temp_set = self.get_temp_set()
        if temp_set is not None:
            data[AZD_TEMP_SET] = temp_set

        return data

    def get_action(self) -> OperationAction:
        """Return DeviceGroup action."""
        action_values: list[OperationAction] = []
        for aidoo in self.aidoos.values():
            action_values += [aidoo.get_action()]
        for zone in self.zones.values():
            action_values += [zone.get_action()]
        if len(action_values) > 0:
            action_cnt = Counter(action_values)
            idle_cnt = action_cnt[OperationAction.IDLE]
            del action_cnt[OperationAction.IDLE]
            del action_cnt[OperationAction.OFF]
            action_mc = action_cnt.most_common(1)
            if len(action_mc) > 0:
                return action_mc[0][0]
            if idle_cnt > 0:
                return OperationAction.IDLE
        return OperationAction.OFF

    def get_active(self) -> bool | None:
        """Return DeviceGroup active status."""
        active_values: list[bool] = []
        for aidoo in self.aidoos.values():
            active = aidoo.get_active()
            if active is not None:
                active_values += [active]
        for zone in self.zones.values():
            active = zone.get_active()
            if active is not None:
                active_values += [active]
        if len(active_values) > 0:
            if True in active_values:
                return True
            if False in active_values:
                return False
        return None

    def add_aidoo(self, aidoo: Aidoo) -> None:
        """Add Aidoo to DeviceGroup."""
        aidoo_id = aidoo.get_id()
        if aidoo_id not in self.aidoos:
            self.aidoos[aidoo_id] = aidoo

    def add_system(self, system: System) -> None:
        """Add System to DeviceGroup."""
        system_id = system.get_id()
        if system_id not in self.systems:
            self.systems[system_id] = system

    def add_zone(self, zone: Zone) -> None:
        """Add Zone to DeviceGroup."""
        zone_id = zone.get_id()
        if zone_id not in self.zones:
            self.zones[zone_id] = zone

    def get_available(self) -> bool:
        """Return DeviceGroup availability status."""
        for aidoo in self.aidoos.values():
            if aidoo.get_available():
                return True
        for zone in self.zones.values():
            if zone.get_available():
                return True
        return False

    def get_devices_num(self) -> int:
        """Return DeviceGroup devices count."""
        return len(self.aidoos) + len(self.zones)

    def get_humidity(self) -> int | None:
        """Return DeviceGroup humidity."""
        humidity: int | float | None
        humidity_values: list[int] = []
        for aidoo in self.aidoos.values():
            humidity = aidoo.get_humidity()
            if humidity is not None:
                humidity_values += [humidity]
        for zone in self.zones.values():
            humidity = zone.get_humidity()
            if humidity is not None:
                humidity_values += [humidity]
        if len(humidity_values) > 0:
            humidity = sum(humidity_values) / len(humidity_values)
            return int(humidity)
        return None

    def get_id(self) -> str:
        """Return DeviceGroup ID."""
        return self.id

    def get_mode(self) -> OperationMode | None:
        """Return DeviceGroup mode."""
        mode_values: list[OperationMode] = []
        for aidoo in self.aidoos.values():
            mode = aidoo.get_mode()
            if mode is not None:
                mode_values += [mode]
        for zone in self.zones.values():
            mode = zone.get_mode()
            if mode is not None:
                mode_values += [mode]
        if len(mode_values) > 0:
            mode_cnt = Counter(mode_values)
            del mode_cnt[OperationMode.STOP]
            mode_mc = mode_cnt.most_common(1)
            if len(mode_mc) > 0:
                return mode_mc[0][0]
        return OperationMode.STOP

    def get_modes(self) -> list[OperationMode] | None:
        """Return DeviceGroup modes."""
        mode_values: list[OperationMode] = []
        for aidoo in self.aidoos.values():
            modes = aidoo.get_modes() or []
            for mode in modes:
                if mode not in mode_values:
                    mode_values += [mode]
        for zone in self.zones.values():
            modes = zone.get_modes() or []
            for mode in modes:
                if mode not in mode_values:
                    mode_values += [mode]
        if len(mode_values) > 0:
            return mode_values
        return None

    def get_name(self) -> str:
        """Return DeviceGroup name."""
        return self.name

    def get_power(self) -> bool | None:
        """Return DeviceGroup power status."""
        power_values: list[bool] = []
        for aidoo in self.aidoos.values():
            power = aidoo.get_power()
            if power is not None:
                power_values += [power]
        for zone in self.zones.values():
            power = zone.get_power()
            if power is not None:
                power_values += [power]
        if len(power_values) > 0:
            if True in power_values:
                return True
            if False in power_values:
                return False
        return None

    def get_temperature(self) -> float | None:
        """Return DeviceGroup temperature."""
        temp_values: list[float] = []
        for aidoo in self.aidoos.values():
            temp = aidoo.get_temperature()
            if temp is not None:
                temp_values += [temp]
        for zone in self.zones.values():
            temp = zone.get_temperature()
            if temp is not None:
                temp_values += [temp]
        if len(temp_values) > 0:
            temp = sum(temp_values) / len(temp_values)
            return round(temp, 1)
        return None

    def get_temp_set(self) -> float | None:
        """Return DeviceGroup setpoint."""
        temp_set_values: list[float] = []
        for aidoo in self.aidoos.values():
            temp_set = aidoo.get_temp_set()
            if temp_set is not None:
                temp_set_values += [temp_set]
        for zone in self.zones.values():
            temp_set = zone.get_temp_set()
            if temp_set is not None:
                temp_set_values += [temp_set]
        if len(temp_set_values) > 0:
            temp_set = sum(temp_set_values) / len(temp_set_values)
            return round(temp_set, 1)
        return None

    def get_temp_set_max(self) -> float | None:
        """Return DeviceGroup max setpoint."""
        temp_set_max_values: list[float] = []
        for aidoo in self.aidoos.values():
            temp_set_max = aidoo.get_temp_set_max()
            if temp_set_max is not None:
                temp_set_max_values += [temp_set_max]
        for zone in self.zones.values():
            temp_set_max = zone.get_temp_set_max()
            if temp_set_max is not None:
                temp_set_max_values += [temp_set_max]
        if len(temp_set_max_values) > 0:
            temp_set_max = sum(temp_set_max_values) / len(temp_set_max_values)
            return round(temp_set_max, 1)
        return None

    def get_temp_set_min(self) -> float | None:
        """Return DeviceGroup min setpoint."""
        temp_set_min_values: list[float] = []
        for aidoo in self.aidoos.values():
            temp_set_min = aidoo.get_temp_set_min()
            if temp_set_min is not None:
                temp_set_min_values += [temp_set_min]
        for zone in self.zones.values():
            temp_set_min = zone.get_temp_set_min()
            if temp_set_min is not None:
                temp_set_min_values += [temp_set_min]
        if len(temp_set_min_values) > 0:
            temp_set_min = sum(temp_set_min_values) / len(temp_set_min_values)
            return round(temp_set_min, 1)
        return None

    def get_temp_step(self) -> float | None:
        """Return DeviceGroup temperature step."""
        return API_DEFAULT_TEMP_STEP

    def set_params(self, params: dict[str, Any]) -> None:
        """Update DeviceGroup parameters from API request."""

        _params: dict[str, Any] = params[API_PARAMS]
        for param, value in _params.items():
            data: dict[str, Any] = {
                API_VALUE: value,
            }
            for aidoo in self.aidoos.values():
                aidoo.set_param(param, data)
            for zone in self.zones.values():
                zone.set_param(param, data)



"""Airzone Cloud API Group."""

from __future__ import annotations

from typing import Any

from .const import API_GROUP_ID, API_NAME, AZD_INSTALLATION
from .device_group import DeviceGroup


class Group(DeviceGroup):
    """Airzone Cloud Group."""

    def __init__(self, inst_id: str, group_data: dict[str, Any]) -> None:
        """Airzone Cloud Group init."""
        super().__init__()

        self.id = str(group_data[API_GROUP_ID])
        self.installation_id = inst_id

        name: str = group_data.get(API_NAME, "")
        if len(name) > 0:
            self.name = name
        else:
            self.name = "Group"

    def data(self) -> dict[str, Any]:
        """Return Group data."""
        data = super().data()

        data[AZD_INSTALLATION] = self.get_installation()

        return data

    def get_installation(self) -> str:
        """Return Installation ID."""
        return self.installation_id



"""Airzone Cloud API common code."""

from __future__ import annotations

from dataclasses import dataclass
from enum import IntEnum, StrEnum
from typing import Any


@dataclass
class ConnectionOptions:
    """Airzone Cloud options for connection."""

    username: str
    password: str
    websockets: bool = False


class AirQualityMode(StrEnum):
    """Airzone Cloud air quality modes."""

    UNKNOWN = "unknown"

    OFF = "off"
    ON = "on"
    AUTO = "auto"

    @classmethod
    def _missing_(cls, value: Any) -> AirQualityMode:
        return cls.UNKNOWN


class OperationAction(IntEnum):
    """Airzone Cloud operation actions."""

    COOLING = 1
    DRYING = 2
    FAN = 3
    HEATING = 4
    IDLE = 5
    OFF = 6


class OperationMode(IntEnum):
    """Airzone Cloud operation modes."""

    STOP = 0
    AUTO = 1
    COOLING = 2
    HEATING = 3
    VENTILATION = 4
    DRY = 5
    EMERGENCY_HEAT = 6
    HEAT_AIR = 7
    HEAT_RADIANT = 8
    HEAT_COMBINED = 9
    COOLING_AIR = 10
    COOLING_RADIANT = 11
    COOLING_COMBINED = 12

    def is_auto(self) -> bool:
        """Return if mode is Auto."""
        return self.value in [self.AUTO]

    def is_cool(self) -> bool:
        """Return if mode is Cool."""
        return self.value in [
            self.COOLING,
            self.COOLING_AIR,
            self.COOLING_COMBINED,
            self.COOLING_RADIANT,
        ]

    def is_dry(self) -> bool:
        """Return if mode is Dry."""
        return self.value in [self.DRY]

    def is_heat(self) -> bool:
        """Return if mode is Heat."""
        return self.value in [
            self.EMERGENCY_HEAT,
            self.HEATING,
            self.HEAT_AIR,
            self.HEAT_COMBINED,
            self.HEAT_RADIANT,
        ]

    def is_stop(self) -> bool:
        """Return if mode is Stop."""
        return self.value in [self.STOP]

    def is_vent(self) -> bool:
        """Return if mode is Ventilation."""
        return self.value in [self.VENTILATION]


class SpeedType(IntEnum):
    """Airzone Cloud speed type."""

    NORMAL = 0
    AIRFLOW = 1


class TemperatureUnit(IntEnum):
    """Airzone Cloud temperature units."""

    CELSIUS = 0
    FAHRENHEIT = 1



"""Airzone Cloud API Installation."""

from __future__ import annotations

from typing import Any

from .const import (
    API_INSTALLATION_ID,
    API_NAME,
    API_WS_IDS,
    AZD_GROUPS,
    AZD_NUM_GROUPS,
    AZD_WEBSERVERS,
)
from .device_group import DeviceGroup
from .group import Group


class Installation(DeviceGroup):
    """Airzone Cloud Installation."""

    def __init__(self, inst_data: dict[str, Any]) -> None:
        """Airzone Cloud Installation init."""
        super().__init__()

        self.groups: dict[str, Group] = {}
        self.id = str(inst_data[API_INSTALLATION_ID])
        self.webservers: list[str] = []

        name: str = inst_data.get(API_NAME, "")
        if len(name) > 0:
            self.name = name
        else:
            self.name = "Installation"

        for ws_id in inst_data[API_WS_IDS]:
            self.webservers += [ws_id]

    def data(self) -> dict[str, Any]:
        """Return Installation data."""
        data = super().data()

        if len(self.groups) > 0:
            data[AZD_GROUPS] = list(self.groups)

        data[AZD_NUM_GROUPS] = self.get_groups_num()
        data[AZD_WEBSERVERS] = self.get_webservers()

        return data

    def add_group(self, group: Group) -> None:
        """Add Group to Installation."""
        group_id = group.get_id()
        if group_id not in self.groups:
            self.groups[group_id] = group

    def get_groups_num(self) -> int:
        """Return Groups count."""
        return len(self.groups)

    def get_id(self) -> str:
        """Return Installation ID."""
        return self.id

    def get_name(self) -> str:
        """Return Installation name."""
        return self.name

    def get_webservers(self) -> list[str]:
        """Return Installation WebServers."""
        return self.webservers



"""Airzone Cloud API."""

from __future__ import annotations

import asyncio
from asyncio import Semaphore
import logging
from typing import Any, cast
import urllib.parse

from aiohttp import ClientConnectorError, ClientResponseError, ClientSession

from .aidoo import Aidoo
from .common import ConnectionOptions, OperationMode
from .const import (
    API_AUTH_LOGIN,
    API_AUTH_REFRESH_TOKEN,
    API_AZ_AIDOO,
    API_AZ_AIDOO_PRO,
    API_AZ_SYSTEM,
    API_AZ_ZONE,
    API_CONFIG,
    API_DEVICE_ID,
    API_DEVICE_TYPE,
    API_DEVICES,
    API_EMAIL,
    API_GROUP,
    API_GROUPS,
    API_INSTALLATION_ID,
    API_INSTALLATIONS,
    API_MODE,
    API_OPTS,
    API_PARAM,
    API_PARAMS,
    API_PASSWORD,
    API_STATUS,
    API_TYPE,
    API_URL,
    API_USER,
    API_USER_LOGOUT,
    API_V1,
    API_VALUE,
    API_WS,
    API_WS_ID,
    AZD_AIDOOS,
    AZD_GROUPS,
    AZD_INSTALLATIONS,
    AZD_SYSTEMS,
    AZD_WEBSERVERS,
    AZD_ZONES,
    HTTP_CALL_TIMEOUT,
    HTTP_MAX_REQUESTS,
    RAW_DEVICES_CONFIG,
    RAW_DEVICES_STATUS,
    RAW_INSTALLATIONS,
    RAW_INSTALLATIONS_LIST,
    RAW_USER,
    RAW_WEBSERVERS,
    WS_INIT_TIMEOUT,
)
from .device import Device
from .entity import EntityUpdate, UpdateType
from .exceptions import (
    AirzoneCloudError,
    APIError,
    AuthError,
    LoginError,
    TokenRefreshError,
    TooManyRequests,
)
from .group import Group
from .installation import Installation
from .system import System
from .token import AirzoneCloudToken
from .webserver import WebServer
from .websockets import AirzoneCloudIWS
from .zone import Zone

_LOGGER = logging.getLogger(__name__)


class AirzoneCloudApi:
    """Airzone Cloud API."""

    def __init__(
        self,
        session: ClientSession | None,
        options: ConnectionOptions,
    ):
        """Airzone Cloud API init."""
        self._api_init_done: bool = False
        self._api_raw_data: dict[str, Any] = {
            RAW_DEVICES_CONFIG: {},
            RAW_DEVICES_STATUS: {},
            RAW_INSTALLATIONS: {},
            RAW_WEBSERVERS: {},
        }
        self._api_semaphore: Semaphore = Semaphore(HTTP_MAX_REQUESTS)
        self.aidoos: dict[str, Aidoo] = {}
        self.devices: dict[str, Device] = {}
        self.groups: dict[str, Group] = {}
        self.installations: dict[str, Installation] = {}
        self.options = options
        self.session: ClientSession | None = session
        self.systems: dict[str, System] = {}
        self.token: AirzoneCloudToken = AirzoneCloudToken()
        self.webservers: dict[str, WebServer] = {}
        self.websockets: dict[str, AirzoneCloudIWS] = {}
        self.zones: dict[str, Zone] = {}

    async def api_request(
        self, method: str, path: str, json: Any | None = None
    ) -> dict[str, Any]:
        """Airzone Cloud API request."""
        _LOGGER.debug("aiohttp request: /%s (params=%s)", path, json)

        if self.session is None:
            session = ClientSession()
        else:
            session = self.session

        await self._api_semaphore.acquire()
        try:
            async with session.request(
                method,
                f"{API_URL}/{path}",
                headers=self.token.headers(),
                json=json,
                raise_for_status=True,
                timeout=HTTP_CALL_TIMEOUT,
            ) as resp:
                resp_json = await resp.json(content_type=None)
        except ClientConnectorError as err:
            raise AirzoneCloudError(err) from err
        except ClientResponseError as err:
            if path.endswith(API_AUTH_LOGIN):
                raise LoginError(err) from err
            if path.endswith(API_AUTH_REFRESH_TOKEN):
                raise TokenRefreshError(err) from err

            if err.status == 400:
                raise APIError(err) from err
            if err.status == 401:
                raise AuthError(err) from err
            if err.status == 429:
                raise TooManyRequests(err) from err

            raise AirzoneCloudError(err) from err
        finally:
            self._api_semaphore.release()
            if self.session is None:
                await session.close()

        _LOGGER.debug("aiohttp response: %s", resp_json)

        return cast(dict[str, Any], resp_json)

    async def api_get_device_config(self, device: Device) -> dict[str, Any]:
        """Request API device config data."""
        dev_id = device.get_id()
        url_id = urllib.parse.quote(dev_id)

        params = {
            API_INSTALLATION_ID: device.get_installation(),
        }
        dev_params = urllib.parse.urlencode(params)

        res = await self.api_request(
            "GET",
            f"{API_V1}/{API_DEVICES}/{url_id}/{API_CONFIG}?{dev_params}",
        )
        self._api_raw_data[RAW_DEVICES_CONFIG][dev_id] = res

        return res

    async def api_get_device_status(self, device: Device) -> dict[str, Any]:
        """Request API device status data."""
        dev_id = device.get_id()
        url_id = urllib.parse.quote(dev_id)

        params = {
            API_INSTALLATION_ID: device.get_installation(),
        }
        dev_params = urllib.parse.urlencode(params)

        res = await self.api_request(
            "GET",
            f"{API_V1}/{API_DEVICES}/{url_id}/{API_STATUS}?{dev_params}",
        )
        self._api_raw_data[RAW_DEVICES_STATUS][dev_id] = res

        return res

    async def api_get_installation(self, installation: Installation) -> dict[str, Any]:
        """Request API installation data."""
        inst_id = installation.get_id()
        url_id = urllib.parse.quote(inst_id)

        res = await self.api_request(
            "GET",
            f"{API_V1}/{API_INSTALLATIONS}/{url_id}",
        )
        self._api_raw_data[RAW_INSTALLATIONS][inst_id] = res

        return res

    async def api_get_installations(self) -> dict[str, Any]:
        """Request API installations data."""
        res = await self.api_request(
            "GET",
            f"{API_V1}/{API_INSTALLATIONS}",
        )
        self._api_raw_data[RAW_INSTALLATIONS_LIST] = res

        return res

    async def api_get_user(self) -> dict[str, Any]:
        """Request API user data."""
        res = await self.api_request(
            "GET",
            f"{API_V1}/{API_USER}",
        )
        self._api_raw_data[RAW_USER] = res

        return res

    async def api_get_webserver(
        self, webserver: WebServer, devices: bool
    ) -> dict[str, Any]:
        """Request API webserver data."""
        ws_id = webserver.get_id()
        url_id = urllib.parse.quote(ws_id)

        params: dict[str, Any] = {
            API_INSTALLATION_ID: webserver.get_installation(),
        }
        if devices:
            params[API_DEVICES] = 1
        ws_params = urllib.parse.urlencode(params)

        res = await self.api_request(
            "GET",
            f"{API_V1}/{API_DEVICES}/{API_WS}/{url_id}/{API_STATUS}?{ws_params}",
        )
        self._api_raw_data[RAW_WEBSERVERS][ws_id] = res

        return res

    async def api_patch_device(
        self, device: Device, json: dict[str, Any]
    ) -> dict[str, Any]:
        """Perform a PATCH request to update device parameters."""
        dev_id = device.get_id()
        url_id = urllib.parse.quote(dev_id)

        return await self.api_request(
            "PATCH",
            f"{API_V1}/{API_DEVICES}/{url_id}",
            json,
        )

    async def api_put_group(self, group: Group, json: dict[str, Any]) -> dict[str, Any]:
        """Perform a PUT request to update group parameters."""
        grp_id = group.get_id()
        grp_url = urllib.parse.quote(grp_id)
        inst_id = group.get_installation()
        inst_url = urllib.parse.quote(inst_id)

        return await self.api_request(
            "PUT",
            f"{API_V1}/{API_INSTALLATIONS}/{inst_url}/{API_GROUP}/{grp_url}",
            json,
        )

    async def api_put_installation(
        self, inst: Installation, json: dict[str, Any]
    ) -> dict[str, Any]:
        """Perform a PUT request to update installation parameters."""
        inst_id = inst.get_id()
        inst_url = urllib.parse.quote(inst_id)

        return await self.api_request(
            "PUT",
            f"{API_V1}/{API_INSTALLATIONS}/{inst_url}",
            json,
        )

    def api_conv_special_mode(
        self, modes: list[OperationMode], mode: OperationMode
    ) -> OperationMode:
        """Convert Home Assistant Operation Mode into its corresponding API value."""
        if mode == OperationMode.COOLING:
            if OperationMode.COOLING_COMBINED in modes:
                mode = OperationMode.COOLING_COMBINED
            elif OperationMode.COOLING_AIR in modes:
                mode = OperationMode.COOLING_AIR
            elif OperationMode.COOLING_RADIANT in modes:
                mode = OperationMode.COOLING_RADIANT
        elif mode == OperationMode.HEATING:
            if OperationMode.HEAT_COMBINED in modes:
                mode = OperationMode.HEAT_COMBINED
            elif OperationMode.HEAT_AIR in modes:
                mode = OperationMode.HEAT_AIR
            elif OperationMode.HEAT_RADIANT in modes:
                mode = OperationMode.HEAT_RADIANT
            elif OperationMode.EMERGENCY_HEAT in modes:
                mode = OperationMode.EMERGENCY_HEAT
        return mode

    async def api_set_device_param(
        self, device: Device, param: str, data: dict[str, Any]
    ) -> None:
        """Set device parameter."""
        if param == API_MODE:
            modes = device.get_modes() or []
            if data[API_VALUE] not in modes:
                data[API_VALUE] = self.api_conv_special_mode(
                    modes, data[API_VALUE]
                ).value

            if isinstance(device, Zone) and not device.get_master():
                # Mode can't be changed on slave zones
                system = device.get_system()
                if system is not None:
                    device = system

        json = {
            API_PARAM: param,
            API_VALUE: data[API_VALUE],
            API_INSTALLATION_ID: device.get_installation(),
        }

        if API_OPTS in data:
            json[API_OPTS] = data[API_OPTS]

        await self.api_patch_device(device, json)

        device.set_param(param, data)

    async def api_set_device_params(
        self, device: Device, params: dict[str, Any]
    ) -> None:
        """Set device parameters."""
        tasks = []

        for param, data in params.items():
            tasks += [self.api_set_device_param(device, param, data)]

        await asyncio.gather(*tasks)

    async def api_set_group_params(self, group: Group, params: dict[str, Any]) -> None:
        """Set group parameters."""
        for key, value in params.get(API_PARAMS, {}).items():
            if key == API_MODE:
                modes = group.get_modes() or []
                if value not in modes:
                    params[API_PARAMS][key] = self.api_conv_special_mode(
                        modes, value
                    ).value

        await self.api_put_group(group, params)

        group.set_params(params)

    async def api_set_installation_params(
        self, inst: Installation, params: dict[str, Any]
    ) -> None:
        """Set installation parameters."""
        for key, value in params.get(API_PARAMS, {}).items():
            if key == API_MODE:
                modes = inst.get_modes() or []
                if value not in modes:
                    params[API_PARAMS][key] = self.api_conv_special_mode(
                        modes, value
                    ).value

        await self.api_put_installation(inst, params)

        inst.set_params(params)

    async def api_set_aidoo_id_params(
        self, aidoo_id: str, params: dict[str, Any]
    ) -> None:
        """Set aidoo parameters."""
        aidoo = self.get_aidoo_id(aidoo_id)
        if aidoo is not None:
            await self.api_set_device_params(aidoo, params)

    async def api_set_device_id_params(
        self, device_id: str, params: dict[str, Any]
    ) -> None:
        """Set device parameters."""
        device = self.get_device_id(device_id)
        if device is not None:
            await self.api_set_device_params(device, params)

    async def api_set_group_id_params(
        self, group_id: str, params: dict[str, Any]
    ) -> None:
        """Set group parameters."""
        group = self.get_group_id(group_id)
        if group is not None:
            await self.api_set_group_params(group, params)

    async def api_set_installation_id_params(
        self, inst_id: str, params: dict[str, Any]
    ) -> None:
        """Set installation parameters."""
        inst = self.get_installation_id(inst_id)
        if inst is not None:
            await self.api_set_installation_params(inst, params)

    async def api_set_system_id_params(
        self, system_id: str, params: dict[str, Any]
    ) -> None:
        """Set system parameters."""
        system = self.get_system_id(system_id)
        if system is not None:
            await self.api_set_device_params(system, params)

    async def api_set_zone_id_params(
        self, zone_id: str, params: dict[str, Any]
    ) -> None:
        """Set zone parameters."""
        zone = self.get_zone_id(zone_id)
        if zone is not None:
            await self.api_set_device_params(zone, params)

    async def login(self) -> None:
        """Perform Airzone Cloud API login."""
        if self.token.is_valid():
            await self.logout()
        resp = await self.api_request(
            "POST",
            f"{API_V1}/{API_AUTH_LOGIN}",
            {
                API_EMAIL: self.options.username,
                API_PASSWORD: self.options.password,
            },
        )
        _LOGGER.debug("login resp: %s", resp)
        self.token.update(resp, False)

    async def logout(self) -> None:
        """Perform Airzone Cloud API logout."""
        for inst_ws in self.websockets.values():
            inst_ws.disconnect()

        try:
            if self.token.is_valid():
                await self.api_request(
                    "GET",
                    f"{API_V1}/{API_USER_LOGOUT}",
                )
        except AirzoneCloudError:
            pass
        finally:
            self.token.clear()

    async def token_refresh(self) -> None:
        """Perform Airzone Cloud API token refresh."""
        if self.token.is_valid():
            resp = await self.api_request(
                "GET",
                self.token.url_refresh(),
            )
            _LOGGER.debug("refresh resp: %s", resp)
            self.token.update(resp, True)

    def raw_data(self) -> dict[str, Any]:
        """Return raw Airzone Cloud API data."""
        return self._api_raw_data

    def data(self) -> dict[str, Any]:
        """Return Airzone Cloud data."""
        data: dict[str, Any] = {}

        if len(self.aidoos) > 0:
            aidoos: dict[str, Any] = {}
            for key, aidoo in self.aidoos.items():
                aidoos[key] = aidoo.data()
            data[AZD_AIDOOS] = aidoos

        if len(self.groups) > 0:
            groups: dict[str, Any] = {}
            for key, group in self.groups.items():
                groups[key] = group.data()
            data[AZD_GROUPS] = groups

        if len(self.installations) > 0:
            installations: dict[str, Any] = {}
            for key, installation in self.installations.items():
                installations[key] = installation.data()
            data[AZD_INSTALLATIONS] = installations

        if len(self.systems) > 0:
            systems: dict[str, Any] = {}
            for key, system in self.systems.items():
                systems[key] = system.data()
            data[AZD_SYSTEMS] = systems

        if len(self.webservers) > 0:
            webservers: dict[str, Any] = {}
            for key, webserver in self.webservers.items():
                webservers[key] = webserver.data()
            data[AZD_WEBSERVERS] = webservers

        if len(self.zones) > 0:
            zones: dict[str, Any] = {}
            for key, zone in self.zones.items():
                zones[key] = zone.data()
            data[AZD_ZONES] = zones

        return data

    def add_aidoo(self, aidoo: Aidoo) -> None:
        """Add Airzone Cloud Aidoo."""
        self.aidoos[aidoo.get_id()] = aidoo
        self.add_device(aidoo)

    def add_device(self, device: Device) -> None:
        """Add Airzone Cloud Device."""
        dev_id = device.get_id()
        if dev_id not in self.devices:
            self.devices[dev_id] = device

    def add_system(self, system: System) -> None:
        """Add Airzone Cloud System."""
        self.systems[system.get_id()] = system
        self.add_device(system)

    def add_zone(self, zone: Zone) -> None:
        """Add Airzone Cloud System."""
        self.zones[zone.get_id()] = zone
        self.add_device(zone)

    def get_aidoo_id(self, aidoo_id: str) -> Aidoo | None:
        """Return Airzone Cloud Aidoo by ID."""
        return self.aidoos.get(aidoo_id)

    def get_device_id(self, dev_id: str | None) -> Device | None:
        """Return Airzone Cloud Device by ID."""
        if dev_id is not None:
            return self.devices.get(dev_id)
        return None

    def get_group_id(self, group_id: str) -> Group | None:
        """Return Airzone Cloud Group by ID."""
        return self.groups.get(group_id)

    def get_installation_id(self, inst_id: str) -> Installation | None:
        """Return Airzone Cloud Installation by ID."""
        return self.installations.get(inst_id)

    def get_system_id(self, sys_id: str) -> System | None:
        """Return Airzone Cloud System by ID."""
        return self.systems.get(sys_id)

    def get_webserver_id(self, ws_id: str | None) -> WebServer | None:
        """Return Airzone Cloud WebServer by ID."""
        if ws_id is not None:
            return self.webservers.get(ws_id)
        return None

    def get_zone_id(self, zone_id: str) -> Zone | None:
        """Return Airzone Cloud Zone by ID."""
        return self.zones.get(zone_id)

    async def list_installations(self) -> list[Installation]:
        """Return Airzone Cloud installations list."""
        inst_list: list[Installation] = []

        inst_data = await self.api_get_installations()
        for inst in inst_data[API_INSTALLATIONS]:
            inst_list += [Installation(inst)]

        return inst_list

    def select_installation(self, inst: Installation) -> None:
        """Select single Airzone Cloud installation."""
        inst_id = inst.get_id()

        if self.options.websockets:
            for _inst in list(self.installations):
                inst_ws = self.websockets.get(_inst)
                if inst_ws is not None and inst_ws.disconnect():
                    self.websockets.pop(_inst)

            if inst_id not in self.websockets:
                self.websockets[inst_id] = AirzoneCloudIWS(self, inst)

        self.installations = {
            inst_id: inst,
        }

        for ws_id in inst.get_webservers():
            if self.get_webserver_id(ws_id) is None:
                self.webservers[ws_id] = WebServer(inst_id, ws_id)

    def set_system_zones_data(self, system: System) -> None:
        """Set slave zones modes from master zone."""
        modes = system.get_modes()
        installation_id = system.get_installation()
        system_num = system.get_system_num()
        system_ws = system.get_webserver()
        for zone in self.zones.values():
            if (
                zone.get_installation() != installation_id
                or zone.get_system_num() != system_num
                or zone.get_webserver() != system_ws
            ):
                continue

            system.add_zone(zone)
            zone.set_system(system)
            if zone.get_master() is False and modes:
                zone.set_modes(modes)

    async def update_aidoo(self, aidoo: Aidoo) -> None:
        """Update Airzone Cloud Zone from API."""
        device_data = await self.api_get_device_status(aidoo)

        update = EntityUpdate(UpdateType.API_FULL, device_data)

        await aidoo.update(update)

    async def update_aidoos(self) -> None:
        """Update all Airzone Cloud Aidoos."""
        tasks = []

        for aidoo in self.aidoos.values():
            tasks += [self.update_aidoo(aidoo)]

        await asyncio.gather(*tasks)

    async def update_installation(self, inst: Installation) -> None:
        """Update Airzone Cloud installation from API."""
        inst_id = inst.get_id()
        installation_data = await self.api_get_installation(inst)
        for group_data in installation_data[API_GROUPS]:
            group = Group(inst_id, group_data)
            inst.add_group(group)
            self.groups[group.get_id()] = group
            for device_data in group_data[API_DEVICES]:
                device_id = device_data[API_DEVICE_ID]
                device_type = device_data[API_TYPE]
                if device_type == API_AZ_ZONE:
                    if self.get_zone_id(device_id) is None:
                        zone = Zone(inst_id, device_data[API_WS_ID], device_data)
                        if zone is not None:
                            self.add_zone(zone)
                            group.add_zone(zone)
                            inst.add_zone(zone)
                elif device_type == API_AZ_SYSTEM:
                    if self.get_system_id(device_id) is None:
                        system = System(inst_id, device_data[API_WS_ID], device_data)
                        if system is not None:
                            self.add_system(system)
                            group.add_system(system)
                            inst.add_system(system)
                elif device_type in [API_AZ_AIDOO, API_AZ_AIDOO_PRO]:
                    if self.get_aidoo_id(device_id) is None:
                        aidoo = Aidoo(inst_id, device_data[API_WS_ID], device_data)
                        if aidoo is not None:
                            self.add_aidoo(aidoo)
                            group.add_aidoo(aidoo)
                            inst.add_aidoo(aidoo)

        inst_ws = self.websockets.get(inst_id)
        if inst_ws is not None:
            inst_ws.connect()

    async def update_installations(self) -> None:
        """Update Airzone Cloud installations from API."""
        installations_data = await self.api_get_installations()
        for installation_data in installations_data[API_INSTALLATIONS]:
            if self.get_installation_id(installation_data[API_INSTALLATION_ID]) is None:
                installation = Installation(installation_data)
                if installation is not None:
                    inst_id = installation.get_id()
                    self.installations[inst_id] = installation
                    for ws_id in installation.get_webservers():
                        if self.get_webserver_id(ws_id) is None:
                            ws = WebServer(inst_id, ws_id)
                            self.webservers[ws_id] = ws

    def get_ws_device_data(self, device: Device) -> dict[str, Any] | None:
        """Get WebSockets device data."""
        inst_id = device.get_installation()
        ws = self.websockets.get(inst_id)
        if ws is not None:
            return ws.get_device_data(device)
        return None

    async def update_system(self, system: System) -> None:
        """Update Airzone Cloud System from API."""
        device_data = await self.api_get_device_status(system)

        update = EntityUpdate(UpdateType.API_FULL, device_data)

        await system.update(update)

    async def update_system_id(self, sys_id: str) -> None:
        """Update Airzone Cloud System by ID."""
        system = self.get_system_id(sys_id)
        if system is not None:
            await self.update_system(system)

    async def update_systems(self) -> None:
        """Update all Airzone Cloud Systems."""
        tasks = []

        for system in self.systems.values():
            tasks += [self.update_system(system)]

        await asyncio.gather(*tasks)

    async def update_systems_zones(self) -> None:
        """Update all Airzone Cloud Systems/Zones."""
        await self.update_systems()
        await self.update_zones()

    async def update_webserver(self, ws: WebServer, devices: bool) -> None:
        """Update Airzone Cloud WebServer from API."""
        ws_data = await self.api_get_webserver(ws, devices)

        update = EntityUpdate(UpdateType.API_FULL, ws_data)

        await ws.update(update)
        if devices:
            ws_id = ws.get_id()
            inst = self.get_installation_id(ws.get_installation())
            for device_data in ws_data[API_DEVICES]:
                device_id = device_data[API_DEVICE_ID]
                device_type = device_data[API_DEVICE_TYPE]
                if device_type == API_AZ_ZONE:
                    if self.get_zone_id(device_id) is None:
                        zone = Zone(ws.get_installation(), ws_id, device_data)
                        if zone is not None:
                            self.add_zone(zone)
                            if inst is not None:
                                inst.add_zone(zone)
                elif device_type == API_AZ_SYSTEM:
                    if self.get_system_id(device_id) is None:
                        system = System(ws.get_installation(), ws_id, device_data)
                        if system is not None:
                            self.add_system(system)
                            if inst is not None:
                                inst.add_system(system)
                elif device_type in [API_AZ_AIDOO, API_AZ_AIDOO_PRO]:
                    if self.get_aidoo_id(device_id) is None:
                        aidoo = Aidoo(ws.get_installation(), ws_id, device_data)
                        if aidoo is not None:
                            self.add_aidoo(aidoo)
                            if inst is not None:
                                inst.add_aidoo(aidoo)

    async def update_webserver_id(self, ws_id: str, devices: bool) -> None:
        """Update Airzone Cloud WebServer by ID."""
        ws = self.get_webserver_id(ws_id)
        if ws is not None:
            await self.update_webserver(ws, devices)

    async def update_webservers(self, devices: bool) -> None:
        """Update all Airzone Cloud WebServers."""
        tasks = []

        for ws in self.webservers.values():
            tasks += [self.update_webserver(ws, devices)]

        await asyncio.gather(*tasks)

    async def update_zone(self, zone: Zone) -> None:
        """Update Airzone Cloud Zone from API."""
        device_data = await self.api_get_device_status(zone)

        update = EntityUpdate(UpdateType.API_FULL, device_data)

        await zone.update(update)

    async def update_zone_id(self, zone_id: str) -> None:
        """Update Airzone Cloud Zone by ID."""
        zone = self.get_zone_id(zone_id)
        if zone is not None:
            await self.update_zone(zone)

    async def update_zones(self) -> None:
        """Update all Airzone Cloud Zones."""
        tasks = []

        for zone in self.zones.values():
            tasks += [self.update_zone(zone)]

        await asyncio.gather(*tasks)

        for system in self.systems.values():
            self.set_system_zones_data(system)

    async def _update_polling(self) -> None:
        """Perform a polling update of Airzone Cloud data."""
        tasks = [
            self.update_webservers(False),
            self.update_systems_zones(),
            self.update_aidoos(),
        ]

        await asyncio.gather(*tasks)
        self._api_init_done = True

    async def _update_websockets(self) -> bool:
        """Perform a websockets update of Airzone Cloud data."""
        if not self.options.websockets:
            return False

        ws_updated = True
        for websocket in self.websockets.values():
            if not websocket.is_connected():
                websocket.connect()
            try:
                await asyncio.wait_for(websocket.state_end.wait(), WS_INIT_TIMEOUT)
            except asyncio.TimeoutError:
                ws_updated = False
        return ws_updated

    async def _update(self) -> None:
        """Update Airzone Cloud data using websockets and fall back to polling."""
        if not self._api_init_done or not await self._update_websockets():
            await self._update_polling()

    async def update(self) -> None:
        """Update all Airzone Cloud data."""

        if self.token.check_refresh():
            try:
                await self.token_refresh()
            except TokenRefreshError:
                await self.login()

        try:
            await self._update()
        except LoginError:
            await self.login()
            await self._update()
